from __future__ import annotations

from typing import Any

from ckanext.selftracking.model.selftracking import SelfTrackingModel
import ckanext.selftracking.utils as tracking_utils


def selftracking_categories():
    return tracking_utils.get_selftracking_categories()


def selftracking_get_data() -> dict[str, Any]:
    count_by_type = SelfTrackingModel.get_tracks_by_types()
    tracks_last_24_hours = SelfTrackingModel.get_tracks_for_last_24_hours()
    track_types_count = SelfTrackingModel.get_tracks_count_for_x_days(30)

    return {
        "count_by_type": count_by_type,
        "tracks_last_24_hours": tracks_last_24_hours,
        "track_types_count": track_types_count,
    }


def selftracking_get_view_data(type: str) -> list[Any]:
    view_data = SelfTrackingModel.get_tracks_per_type(type)
    return view_data



from __future__ import annotations

from typing import Any

from ckan.plugins import Interface


class ISelftracking(Interface):
    """Implement custom Selftracking response modification."""

    def selftracking_categories(self, categories: list[Any]) -> list[Any]:
        """Return list of categories.
        This method is called before returning the categories list.
        Useful if you want to add additional custom categories.
        :returns: list
        :rtype: list

        """

        return categories



from __future__ import annotations

from typing import Any

import ckan.plugins.toolkit as tk


SELFTRACKING_REDIS_PREFIX = "ckan.selftracking.redis_prefix"
SELFTRACKING_CATEGORIES = "ckan.selftracking.categories"
SELFTRACKING_REDIS_BATCH_SIZE = "ckan.selftracking.redis_batch_size"
SELFTRACKING_TRACK_TYPE_COLOR_PREFIX = "ckan.selftracking.type_color."


def selftracking_get_redis_prefix():
    return tk.config.get(SELFTRACKING_REDIS_PREFIX)


def selftracking_get_categories_list():
    return tk.config.get(SELFTRACKING_CATEGORIES)


def selftracking_redis_batch_size():
    return tk.config.get(SELFTRACKING_REDIS_BATCH_SIZE)


def selftracking_type_colors() -> dict[str, dict[str, Any]]:
    colors = {}
    prefix_len = len(SELFTRACKING_TRACK_TYPE_COLOR_PREFIX)
    options = {
        k: v
        for k, v in tk.config.items()
        if k.startswith(SELFTRACKING_TRACK_TYPE_COLOR_PREFIX)
    }
    for k, v in options.items():
        try:
            type = k[prefix_len:]
            colors[type.replace("_", " ")] = v
        except ValueError:
            continue

    return colors


def selftracking_type_color(type: str) -> str:
    color = tk.config.get(
        SELFTRACKING_TRACK_TYPE_COLOR_PREFIX + type.replace(" ", "_"), ""
    )
    return color



from __future__ import annotations

import ckan.plugins as plugins
import ckan.plugins.toolkit as tk
from ckan.common import CKANConfig
import ckan.types as types

from .middleware import track_activity
from .cli import get_commands


@tk.blanket.config_declarations
@tk.blanket.actions
@tk.blanket.auth_functions
@tk.blanket.cli(get_commands)
@tk.blanket.helpers
class SelftrackingPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IMiddleware, inherit=True)

    # IConfigurer
    def update_config(self, config_: CKANConfig):
        tk.add_template_directory(config_, "templates")
        tk.add_public_directory(config_, "public")
        tk.add_resource("assets", "selftracking")

    # IMiddleware
    def make_middleware(self, app: types.CKANApp, _) -> types.CKANApp:
        app.after_request(track_activity)
        return app



import click

import ckan.plugins.toolkit as tk


def get_commands():
    return [selftracking]


@click.group()
def selftracking():
    """selftracking management commands."""


@selftracking.command()
def store_tracking():
    """Store tracking that being gathered in Redis"""
    tk.get_action("selftracking_store_tracks")({"ignore_auth": True}, {})



from __future__ import annotations

import json
import logging
from typing import Any
from datetime import datetime, timezone

import ckan.plugins as p
from ckan.lib.redis import connect_to_redis, Redis

from ckanext.selftracking import interfaces
import ckanext.selftracking.config as s_config
from ckanext.selftracking.model.selftracking import SelfTrackingModel


SELFTRACKING_DEFAULT_CATEGORIES = [
    {
        "key": "main",
        "label": "Main",
        "snippet": "/selftracking/snippets/selftracking_main.html",
    },
    {
        "key": "page_view",
        "label": "Page View",
        "snippet": "/selftracking/snippets/selftracking_page_view.html",
    },
    {
        "key": "api_view",
        "label": "API View",
        "snippet": "/selftracking/snippets/selftracking_api_view.html",
    },
]
log = logging.getLogger(__name__)


def selftracking_add_track_to_redis(data_dict: dict[str, Any]) -> None:
    redis: Redis = connect_to_redis()

    try:
        key = s_config.selftracking_get_redis_prefix() + "_selftracking_tracks"
        redis.rpush(key, json.dumps(data_dict))
    except (ValueError, TypeError):
        log.error("Cannot store track in Redis.")

    return


def selftracking_store_tracks_in_db():
    batch_size = s_config.selftracking_redis_batch_size()
    redis: Redis = connect_to_redis()
    pipe = redis.pipeline()
    key = s_config.selftracking_get_redis_prefix() + "_selftracking_tracks"
    pipe.lrange(key, 0, batch_size - 1)
    pipe.ltrim(key, batch_size, -1)
    batch, _ = pipe.execute()
    for b in batch:
        # For testing big amount
        # for i in range(0, 100000):
        try:
            data = json.loads(b)
            try:
                track_time = data.get("track_time")
                if track_time:
                    track_time = datetime.fromtimestamp(
                        data["track_time"], tz=timezone.utc
                    )
                    data["track_time"] = track_time
            except Exception:
                pass
            SelfTrackingModel.create(data)
        except TypeError:
            log.error("Couldn't write track item to DB.")


def get_categories_list() -> list[dict[str, Any]]:
    categories = SELFTRACKING_DEFAULT_CATEGORIES

    # categories modification
    for item in p.PluginImplementations(interfaces.ISelftracking):
        item.selftracking_categories(categories)

    return categories


def get_selftracking_categories() -> list[dict[str, Any]]:
    categories = s_config.selftracking_get_categories_list()

    categories_list = [
        category for category in get_categories_list() if category["key"] in categories
    ]

    return categories_list



"""Add selftracking table

Revision ID: 1026bee50450
Revises:
Create Date: 2025-09-13 10:46:44.365861

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = "1026bee50450"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "selftracking",
        sa.Column("id", sa.Integer, primary_key=True, unique=True, autoincrement=True),
        sa.Column("path", sa.String, nullable=False),
        sa.Column("user", sa.String, nullable=True),
        sa.Column("type", sa.String, nullable=False),
        sa.Column("track_time", sa.DateTime, default=sa.TIMESTAMP),
        sa.Column("extras", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    )
    op.create_index("idx_selftracking_path", "selftracking", ["path"])
    op.create_index("idx_selftracking_track_time", "selftracking", ["track_time"])


def downgrade():
    op.drop_index("idx_selftracking_track_time", table_name="selftracking")
    op.drop_index("idx_selftracking_path", table_name="selftracking")

    op.drop_table("selftracking")



from __future__ import annotations

from ckan.types import AuthResult, Context, DataDict


def selftracking_store_tracks(context: Context, data_dict: DataDict) -> AuthResult:
    return {"success": False}


def selftracking_send_track_for_queue(
    context: Context, data_dict: DataDict
) -> AuthResult:
    return {"success": True}



from __future__ import annotations

from datetime import datetime
import logging

import ckan.plugins.toolkit as tk
import ckan.types as types

import ckanext.selftracking.utils as s_utils


log = logging.getLogger(__name__)


def selftracking_store_tracks(context: types.Context, data_dict: types.DataDict):
    tk.check_access("selftracking_store_tracks", context, data_dict)

    s_utils.selftracking_store_tracks_in_db()

    return


def selftracking_send_track_for_queue(
    context: types.Context, data_dict: types.DataDict
):
    tk.check_access("selftracking_send_track_for_queue", context, data_dict)

    user = tk.current_user
    data = {
        "path": data_dict.get("path", ""),
        "user": user.id if user.is_authenticated else "anonymous",
        "type": data_dict.get("type", "missing type"),
        "track_time": datetime.utcnow().timestamp(),
    }

    s_utils.selftracking_add_track_to_redis(data)

    return



from __future__ import annotations

from datetime import datetime, timedelta, timezone
import json
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableDict
from typing import Any, List
import hashlib

import ckan.model as model
import ckan.plugins.toolkit as tk

import ckanext.selftracking.config as tracking_config


class SelfTrackingModel(tk.BaseModel):  # type: ignore
    __tablename__ = "selftracking"

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    path = sa.Column(sa.Text, nullable=False)
    user = sa.Column(sa.String, nullable=False)
    type = sa.Column(sa.String, nullable=False)
    track_time = sa.Column(sa.DateTime, default=datetime.utcnow)
    extras = sa.Column(MutableDict.as_mutable(JSONB))

    @classmethod
    def get_by_id(cls: type[SelfTrackingModel], id: str) -> SelfTrackingModel | None:
        return model.Session.query(cls).filter(cls.id == id).first()

    @classmethod
    def get_by_path(
        cls: type[SelfTrackingModel], alias: str
    ) -> SelfTrackingModel | None:
        return model.Session.query(cls).filter(cls.alias == alias).first()

    @classmethod
    def get_by_type(cls: type[SelfTrackingModel], type: str) -> list[SelfTrackingModel]:
        return (
            model.Session.query(cls)
            .filter(cls.type == type)
            .order_by(cls.track_time.desc())
            .all()
        )

    @classmethod
    def get_all(cls: type[SelfTrackingModel]) -> list[SelfTrackingModel]:
        return model.Session.query(cls).order_by(cls.track_time.desc()).all()

    @classmethod
    def create(
        cls: type[SelfTrackingModel], data_dict: dict[str, Any]
    ) -> SelfTrackingModel:
        selftrack = cls(**data_dict)

        model.Session.add(selftrack)
        model.Session.commit()

        return selftrack

    def delete(self) -> None:
        model.Session().autoflush = False
        model.Session.delete(self)
        model.Session.commit()

    def update(self, data_dict: dict[str, Any]) -> None:
        for key, value in data_dict.items():
            setattr(self, key, value)
        model.Session.commit()

    @classmethod
    def color_from_type(cls: type[SelfTrackingModel], type_: str) -> str:
        color = tracking_config.selftracking_type_color(type_)

        if not color:
            h = hashlib.md5(type_.encode()).hexdigest()
            return "#" + h[-6:]

        return color

    @classmethod
    def get_tracks_by_types(cls: type[SelfTrackingModel]) -> list[dict[str, Any]]:
        results = (
            model.Session.query(cls.type, sa.func.count(cls.id))
            .group_by(cls.type)
            .all()
        )

        results = [
            {"type": t[0], "count": t[1], "color": cls.color_from_type(t[0])}
            for t in results
        ]

        return results

    @classmethod
    def get_tracks_by_type_and_days(
        cls: type[SelfTrackingModel], type: str, days: int
    ) -> dict[str, Any]:
        since = datetime.utcnow() - timedelta(days=days)
        results = (
            model.Session.query(
                sa.func.date_trunc("day", cls.track_time).label("day"),
                sa.func.count().label("count"),
            )
            .filter(cls.type == type)
            .filter(cls.track_time >= since)
            .group_by(sa.func.date_trunc("day", cls.track_time))
            .order_by("day")
            .all()
        )

        labels = [r.day.strftime("%Y-%m-%d") for r in results]
        values = [r.count for r in results]

        data = {"labels": labels, "values": values}

        return {"raw": data, "json": json.dumps(data)}

    @classmethod
    def get_tracks_count_for_x_days(
        cls: type[SelfTrackingModel], days: int
    ) -> dict[Any, Any]:
        since = datetime.utcnow() - timedelta(days=days)
        results = (
            model.Session.query(
                cls.type,
                sa.func.date_trunc("day", cls.track_time).label("day"),
                sa.func.count().label("count"),
            )
            .filter(cls.track_time >= since)
            .group_by(cls.type, sa.func.date_trunc("day", cls.track_time))
            .order_by(cls.type, "day")
            .all()
        )

        data = {}
        for r in results:
            type_ = r.type
            if type_ not in data:
                data[type_] = {
                    "labels": [],
                    "values": [],
                    "color": cls.color_from_type(type_),
                }
            data[type_]["labels"].append(r.day.strftime("%Y-%m-%d"))
            data[type_]["values"].append(r.count)

        return data

    @classmethod
    def get_tracks_per_type(
        cls: type[SelfTrackingModel], type: str, t_days: int | None = None
    ) -> List[Any]:

        q = model.Session.query(cls.path, sa.func.count().label("count")).filter(
            cls.type == type
        )

        if t_days:
            since = datetime.utcnow() - timedelta(days=t_days)
            q = q.filter(cls.track_time >= since)

        q = q.group_by(cls.path).order_by(sa.func.count().desc()).all()

        return q

    @classmethod
    def get_tracks_for_last_24_hours(cls: type[SelfTrackingModel]) -> dict[str, Any]:
        end_time = datetime.now(timezone.utc).replace(minute=0, second=0, microsecond=0)
        start_time = end_time - timedelta(hours=23)
        results = (
            model.Session.query(
                cls.type,
                sa.func.date_trunc("hour", cls.track_time).label("hour"),
                sa.func.count().label("count"),
            )
            .filter(cls.track_time >= start_time, cls.track_time <= end_time)
            .group_by(cls.type, "hour")
            .order_by(cls.type, "hour")
            .all()
        )

        data = {key: {} for key, _, _ in results}
        for event_type, hour, count in results:
            local_hour = hour.replace(tzinfo=timezone.utc).astimezone()
            data[event_type][local_hour.strftime("%H:%M")] = count

        labels = [
            (start_time + timedelta(hours=i)).astimezone().strftime("%H:%M")
            for i in range(24)
        ]

        items = {}
        for event_type in data.keys():
            items[event_type] = {
                "label": event_type,
                "data": [data[event_type].get(hour, 0) for hour in labels],
                "color": cls.color_from_type(event_type),
            }

        result = {
            "items": items,
            "labels": labels,
        }
        return result



from __future__ import annotations

from datetime import datetime

import ckan.plugins.toolkit as tk
import ckan.types as types

from ckanext.selftracking.utils import selftracking_add_track_to_redis


def track_activity(response: types.Response) -> types.Response:
    """Add activity to Redis list.

    Adds activity information to Redis list,
    which will be then added to DB.
    """
    path = tk.request.path
    if response.content_type == "text/html; charset=utf-8":
        ignore_paths = [
            "/api/i18n",
            "/favicon.ico",
        ]
        ignore_extensions = [".js", ".css"]

        if not [p for p in ignore_paths if p in path] and not [
            p for p in ignore_extensions if path.endswith(p)
        ]:
            user = tk.current_user
            data = {
                "path": path,
                "user": user.id if user.is_authenticated else "anonymous",
                "type": "page view",
                "track_time": datetime.utcnow().timestamp(),
            }

            selftracking_add_track_to_redis(data)

    elif response.content_type == "application/json;charset=utf-8" and path.startswith(
        "/api/action/"
    ):
        user = tk.current_user
        data = {
            "path": path,
            "user": user.id if user.is_authenticated else "anonymous",
            "type": "api view",
            "track_time": datetime.utcnow().timestamp(),
        }
        selftracking_add_track_to_redis(data)

    return response
